@@ -1,7 +1,7 @@
 package Complete::Util;
 
-our $DATE = '2014-12-24'; # DATE
-our $VERSION = '0.18'; # VERSION
+our $DATE = '2014-12-25'; # DATE
+our $VERSION = '0.19'; # VERSION
 
 use 5.010001;
 use strict;
@@ -240,6 +240,12 @@ $SPEC{complete_file} = {
             summary => 'Case-insensitive matching',
             schema  => 'bool',
         },
+        map_case => {
+            schema  => 'bool',
+        },
+        exp_im_path => {
+            schema  => 'bool',
+        },
         filter => {
             summary => 'Only return items matching this filter',
             description => <<'_',
@@ -267,33 +273,48 @@ _
     },
 };
 sub complete_file {
+    require Complete::Path;
     require File::Glob;
 
     my %args   = @_;
     my $word   = $args{word} // "";
-    my $ci     = $args{ci} // $Complete::OPT_CI;
+    my $ci          = $args{ci} // $Complete::OPT_CI;
+    my $map_case    = $args{map_case} // $Complete::OPT_MAP_CASE;
+    my $exp_im_path = $args{exp_im_path} // $Complete::OPT_EXP_IM_PATH;
     my $filter = $args{filter};
 
     # if word is starts with "~/" or "~foo/" replace it temporarily with user's
     # name (so we can restore it back at the end). this is to mimic bash
     # support. note that bash does not support case-insensitivity for "foo".
-    my $res_prefix; # to be added again to final result
-    my $search_prefix = ''; # to be added when searching
-    my $res_num_remove = 0;
+    my $result_prefix;
+    my $starting_path = '.';
     if ($word =~ s!\A(~[^/]*)/!!) {
-        $res_prefix = "$1/";
+        $result_prefix = "$1/";
         my @dir = File::Glob::glob($1); # glob will expand ~foo to /home/foo
         return [] unless @dir;
-        $search_prefix = $dir[0] =~ m!/\z! ? $dir[0] : "$dir[0]/";
-        $res_num_remove = length($search_prefix);
+        $starting_path = $dir[0];
     } elsif ($word =~ s!\A((?:\.\.?/+)+|/+)!!) {
-        $search_prefix = $1;
-        $search_prefix =~ s#/+\z## unless $search_prefix =~ m!\A/!;
-    } else {
-        $search_prefix = '';
+        # just an optimization to skip sequences of '../'
+        $starting_path = $1;
+        $result_prefix = $1;
+        $starting_path =~ s#/+\z## unless $starting_path =~ m!\A/!;
     }
 
-    # prepare filter sub
+    # prepare list_func
+    my $list = sub {
+        my ($path, $intdir, $isint) = @_;
+        opendir my($dh), $path or return undef;
+        my @res;
+        for (sort readdir $dh) {
+            # skip . and .. if leaf is empty, like in bash
+            next if ($_ eq '.' || $_ eq '..') && $intdir eq '';
+            next if $isint && !(-d "$path/$_");
+            push @res, $_;
+        }
+        \@res;
+    };
+
+    # prepare filter_func
     if ($filter && !ref($filter)) {
         my @seqs = split /\s*\|\s*/, $filter;
         $filter = sub {
@@ -321,91 +342,15 @@ sub complete_file {
         };
     }
 
-    # split word by dirs, as we want to dig level by level (needed when doing
-    # case-insensitive search on a case-sensitive fs). XXX we should optimize
-    # and not split .. or .
-    my @intermediate_dirs = split m!/+!, $word;
-    @intermediate_dirs = ('') if !@intermediate_dirs;
-    push @intermediate_dirs, '' if $word =~ m!/\z!;
-
-    # extract leaf path, because this one is treated differently
-    my $leaf = pop @intermediate_dirs;
-    @intermediate_dirs = ('') if !@intermediate_dirs;
-
-    #say "D:search_prefix=<$search_prefix>";
-    #say "D:intermediate_dirs=[",join(", ", map{"<$_>"} @intermediate_dirs),"]";
-    #say "D:leaf=<$leaf>";
-
-    # candidate for intermediate paths. when doing case-insensitive search,
-    # there maybe multiple candidate paths for each dir, for example if
-    # word='../foo/s' and there is '../foo/Surya', '../Foo/sri', '../FOO/SUPER'
-    # then candidate paths would be ['../foo', '../Foo', '../FOO'] and the
-    # filename should be searched inside all those dirs. everytime we drill down
-    # to deeper subdirectories, we adjust this list.
-    my @candidate_paths;
-
-    for my $i (0..$#intermediate_dirs) {
-        my $intdir = $intermediate_dirs[$i];
-        my @dirs;
-        if ($i == 0) {
-            # first path elem, we search search_prefix first since
-            # candidate_paths is still empty.
-            @dirs = ($search_prefix);
-        } else {
-            # subsequent path elem, we search all candidate_paths
-            @dirs = @candidate_paths;
-        }
-
-        if ($i == $#intermediate_dirs && $intdir eq '') {
-            @candidate_paths = @dirs;
-            last;
-        }
-
-        my @new_candidate_paths;
-        for my $dir (@dirs) {
-            #say "D:  intdir opendir($dir)";
-            opendir my($dh), ($dir eq '' ? '.' : $dir) or next;
-            # check if the deeper level is a directory with the expected name
-            my $re = $ci ? qr/\A\Q$intdir\E\z/i : qr/\A\Q$intdir\E\z/;
-            #say "D:  re=$re";
-            for (sort readdir $dh) {
-                #say "D:  $_";
-                next unless $_ =~ $re;
-                # skip . and .. if leaf is empty, like in bash
-                next if ($_ eq '.' || $_ eq '..') && $intdir eq '';
-                my $p = $dir =~ m!\A\z|/\z! ? "$dir$_" : "$dir/$_";
-                next unless -d $p;
-                push @new_candidate_paths, $p;
-            }
-        }
-        #say "D:  candidate_paths=[",join(", ", map{"<$_>"} @new_candidate_paths),"]";
-        return [] unless @new_candidate_paths;
-        @candidate_paths = @new_candidate_paths;
-    }
-
-    my @res;
-    for my $dir (@candidate_paths) {
-        #say "D:opendir($dir)";
-        opendir my($dh), ($dir eq '' ? '.' : $dir) or next;
-        my $re = $ci ? qr/\A\Q$leaf/i : qr/\A\Q$leaf/;
-        #say "D:re=$re";
-        for (sort readdir $dh) {
-            next unless $_ =~ $re;
-            # skip . and .. if leaf is empty, like in bash
-            next if ($_ eq '.' || $_ eq '..') && $leaf eq '';
-            my $p = $dir =~ m!\A\z|/\z! ? "$dir$_" : "$dir/$_";
-            next if $filter && !$filter->($p);
-
-            # process into final result
-            substr($p, 0, $res_num_remove) = '' if $res_num_remove;
-            $p = "$res_prefix$p" if length($res_prefix);
-            $p .= "/" if -d $p;
-
-            push @res, $p;
-        }
-    }
-
-    \@res;
+    Complete::Path::complete_path(
+        word => $word,
+        ci => $ci, map_case => $map_case, exp_im_path => $exp_im_path,
+        list_func => $list,
+        is_dir_func => sub { -d $_[0] },
+        filter_func => $filter,
+        starting_path => $starting_path,
+        result_prefix => $result_prefix,
+    );
 }
 
 $SPEC{combine_answers} = {
@@ -506,7 +451,7 @@ Complete::Util - General completion routine
 
 =head1 VERSION
 
-This document describes version 0.18 of Complete::Util (from Perl distribution Complete-Util), released on 2014-12-24.
+This document describes version 0.19 of Complete::Util (from Perl distribution Complete-Util), released on 2014-12-25.
 
 =head1 DESCRIPTION
 
@@ -621,6 +566,8 @@ Arguments ('*' denotes required arguments):
 
 Case-insensitive matching.
 
+=item * B<exp_im_path> => I<bool>
+
 =item * B<filter> => I<code|str>
 
 Only return items matching this filter.
@@ -638,6 +585,8 @@ For code filter, you supply a coderef. The coderef will be called for each item
 with these arguments: C<$name>. It should return true if it wants the item to be
 included.
 
+=item * B<map_case> => I<bool>
+
 =item * B<word> => I<str> (default: "")
 
 =back
@@ -732,7 +681,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Complete-U
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/perlancar/perl-Complete-Util>.
+Source repository is at L<https://github.com/sharyanto/perl-SHARYANTO-Complete-Util>.
 
 =head1 BUGS
 
